/*
 * precompile.cxx
 *
 * PWLib application source file for ThreadSafe
 *
 * Precompiled header generation file.
 *
 * Copyright 2002 Equivalence
 *
 * $Revision: 20385 $
 * $Author: rjongbloed $
 * $Date: 2008-06-04 10:40:38 +0000 (Wed, 04 Jun 2008) $
 */

#include <ptlib.h>


// End of File ///////////////////////////////////////////////////////////////
