/*
 * semaphor.h
 *
 * Thread synchronisation semaphore class.
 *
 * Portable Windows Library
 *
 * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
 *
 * Copyright (c) 1999 ISDN Communications Ltd.
 */

#ifndef _PSEMAPHORE

#ifndef __NUCLEUS_MNT__
#pragma interface
#endif

#include <ptlib/thread.h>

#include "../../semaphor.h"

  Semaphore * pNucleusSemaphore;
};

#endif
