mkdir backup
move ptlibd_2003.dtf backup\
copy ..\..\..\lib\debug\ptlibd.def ptlibd_2003.dtf
move ptlib_2003.dtf backup\
copy ..\..\..\lib\release\ptlib.def ptlib_2003.dtf
move ptlibn_2003.dtf backup\
copy "..\..\..\lib\No Trace\ptlibn.def" ptlibn_2003.dtf
