//{{NO_DEPENDENCIES}}
// Microsoft Developer Studio generated include file.
// Used by uninstall.rc
//
#define IDS_NO_WINSOCK                  2
#define IDS_ERROR_DELETING_FILE         3
#define IDS_ERROR_DELETING_DIR          4
#define IDS_STILL_RUNNING               5
#define IDS_STOP_FAILED                 6

// Next default values for new objects
// 
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NEXT_RESOURCE_VALUE        101
#define _APS_NEXT_COMMAND_VALUE         40001
#define _APS_NEXT_CONTROL_VALUE         1000
#define _APS_NEXT_SYMED_VALUE           104
#endif
#endif
