mkdir backup
move ptlibd_2005.dtf backup\
copy ..\..\..\lib\debug\ptlibd.def ptlibd_2005.dtf
move ptlib_2005.dtf backup\
copy ..\..\..\lib\release\ptlib.def ptlib_2005.dtf
move ptlibn_2005.dtf backup\
copy "..\..\..\lib\No Trace\ptlibn.def" ptlibn_2005.dtf
move ptlibd_2005_wm.dtf backup\
copy ..\..\..\lib\wm5ppc\debug\ptlibd.def ptlibd_2005_wm.dtf
move ptlib_2005_wm.dtf backup\
copy ..\..\..\lib\wm5ppc\release\ptlib.def ptlib_2005_wm.dtf
move ptlibn_2005_wm.dtf backup\
copy "..\..\..\lib\wm5ppc\No Trace\ptlibn.def" ptlibn_2005_wm.dtf
move ptlibd_2005_wm6.dtf backup\
copy ..\..\..\lib\wm6pro\debug\ptlibd.def ptlibd_2005_wm6.dtf
move ptlib_2005_wm6.dtf backup\
copy ..\..\..\lib\wm6pro\release\ptlib.def ptlib_2005_wm6.dtf
move ptlibn_2005_wm6.dtf backup\
copy "..\..\..\lib\wm6pro\No Trace\ptlibn.def" ptlibn_2005_wm6.dtf
pause